/*
 Copyright (c) 2012, 2013 G. Bartsch

 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
 files (the "Software"), to deal in the Software without
 restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following
 conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.

 Credits go to:

 arduino-tvout Copyright (c) 2010 Myles Metzer 
 http://code.google.com/p/arduino-tvout/source/browse/TVout.cpp

 AVR Videogenerator, 40x25 Zeichen, nur 60% CPU Auslastung ! by Benedikt
 http://www.mikrocontroller.net/topic/53141#413260

 Simple VGA/Video Adapter von Ibragimov Maxim Rafikovich
 http://www.serasidis.gr/circuits/AVR_VGA/avr_vga.htm
*/

#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "uart/uart.h"

#include "tvgfx/tvgfx.h"
#include "tvfonts/allfonts.h"
#include "img/cloudy.h"
#include "img/mostly_sunny.h"
#include "img/partly_cloudy.h"
#include "img/rain.h"
#include "img/snow.h"
#include "img/sunny.h"
#include "img/thunderstorm.h"

/*
 * PA7 OUT Video PWR Relay (optional)
 *
 * PD0 IN  RXD0 
 * PD1 OUT TXD0 
 * PD2 
 * PD3 OUT Video (TXD1 / MOSI UART1)
 * PD4 
 * PD5 OUT Sync (OC1A)
 * PD6
 * PD7 
 *
 */

#define ENABLE_DEMO

#define UART_BAUD_RATE 38400

#define UART_BAUD_SELECT(baudRate) ((F_CPU)/((baudRate)*16l)-1)

void draw_shuttle(void);

static void init (void) {

	// set up port directions

	// Video PWR RELAY
	DDRA = _BV(7);

	DDRB = 0;
	DDRC = 0;

	//  Sync output (OCR1B), TX0 (UART0), TX1 (Video via SPI UART1)
	DDRD = _BV(5) | _BV(1) | _BV(3) ;

	tvgfx_init();

	// UART (debug output)

	UBRR0 = UART_BAUD_SELECT(UART_BAUD_RATE);

	UCSR0C = 1<<UCSZ01^1<<UCSZ00            // 8 Bit
#ifdef URSEL0
		^1<<URSEL0                      // if UCSR0C shared with UBRR0H
#endif
		;
	UCSR0B = 1<<RXEN0^1<<TXEN1;             // enable RX, TX
	UCSR0A = 0 ;                            // 1x speed

	// video monitor power on	
	PORTA |= (1<<7);

}

#ifdef ENABLE_DEMO
#ifdef TVGFX_FASTFONTS
static uint16_t fontdemo (uint8_t *font, char *fontname, uint16_t ypos) {
#else
static uint16_t fontdemo (uint32_t font, char *fontname, uint16_t ypos) {
#endif

	uint16_t xpos = 5;

	tvgfx_setfont (font);

	xpos += tvgfx_puts(xpos, ypos, fontname);
	xpos += tvgfx_puts(xpos, ypos,": ");
	tvgfx_puts(xpos, ypos, "ABCDEFGHIJKLM");

	ypos += tvgfx_getfontheight();
	tvgfx_puts(xpos, ypos, "abcdefghijklm");

	ypos += tvgfx_getfontheight();
	tvgfx_puts(xpos, ypos, "1234567890!@#$");

	ypos += tvgfx_getfontheight();

	return ypos;
}

static void h9k (char *t, char *st) {
	TVGFX_SETFONT (waukeganeb34);
	tvgfx_puts(150, 100, t);
	TVGFX_SETFONT (waukeganeb26);
	tvgfx_puts (50, 50, st);
}
#endif

int main (void) {

	init();

	sei();

	uart_printf ("Hello, world!\r\n");

#ifdef ENABLE_DEMO

	// cycle through demo screens

	uint8_t demo = 0;

	while (1) {

		tvgfx_clr();
		//tvgfx_rect (0, 0, XSize*8-1, YSize-1, TVGFX_BLACK, TVGFX_BLACK);

		switch (demo) {

			case 0:
				h9k ("W  T  R", "ATA A8-12");
				break;

			case 1:
				TVGFX_SETFONT (waukeganeb26);
				tvgfx_puts(0,85,"FORECAST");
				uint16_t w = pgm_read_byte(thunderstorm);
				uint16_t l = pgm_read_byte(thunderstorm+1);

				tvgfx_bitmap (XSize*8 - w-5, 5, FAR(thunderstorm), 2, w, l);
				TVGFX_SETFONT (waukeganeb34);
				tvgfx_puts(255, 22,"18 C");

				tvgfx_bitmap (33, 110, FAR(cloudy), 2, w, l);
				tvgfx_bitmap (169, 110, FAR(mostly_sunny), 2, w, l);
				tvgfx_bitmap (305, 110, FAR(sunny), 2, w, l);

				TVGFX_SETFONT (terminus16);
				tvgfx_puts(38, 190, "FRI: 27 C");
				tvgfx_puts(174, 190, "SAT: 14 C");
				tvgfx_puts(310, 190, "SUN: 12 C");


				break;

			case 2:
				h9k ("V  E  C", "URN BC-EJ");
				break;

			case 3:
				// vector gfx test

				TVGFX_SETFONT (terminus16);
				tvgfx_puts (0, 0,"VECTOR GFX TEST");

				tvgfx_set_thickness(3);
				tvgfx_set_foreground(TVGFX_WHITE);
				tvgfx_set_background(TVGFX_WHITE);

				draw_shuttle();
				break;

			case 4:
				h9k ("C  O  N", "TXN AC-XX");
				break;

			case 5:
				TVGFX_SETFONT (waukeganeb26);
				tvgfx_puts(0,10,"CONSOLE");
				TVGFX_SETFONT (terminus16);
				tvgfx_puts(0, 60,"DESCRIPTION");
				tvgfx_puts(0,100,"commercial towing vehicle 'The Nostromo'");
				tvgfx_puts(0,140,"crew:   seven");
				tvgfx_puts(0,160,"cargo:  refinery processing");
				tvgfx_puts(0,180,"        20,000,000 tons of mineral ore");
				tvgfx_puts(0,200,"course: returning to earth");
				break;

			case 6:
				h9k ("F  O  N", "GFX A1-YZ");
				break;

			case 7:
				tvgfx_rect (0, 0, XSize*8-1, YSize-1, TVGFX_WHITE, TVGFX_TRANS);
				TVGFX_SETFONT (waukeganeb26);
				tvgfx_puts(140, 3,"FONTS 1");

				uint16_t ypos = 30;
				ypos = fontdemo (FAR(tvout6), "tvout6", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(tvout8), "tvout8", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(tamsynr12), "tamsynr9", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(tamsynr12), "tamsynr12", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(terminus16), "terminus16", ypos);

				break;

			case 8:
				tvgfx_rect (0, 0, XSize*8-1, YSize-1, TVGFX_WHITE, TVGFX_TRANS);
				TVGFX_SETFONT (waukeganeb26);
				tvgfx_puts(140, 3,"FONTS 2");

				ypos = 30;
				ypos = fontdemo (FAR(angle16), "angle16", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(spangle16), "spangle16", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(medieval16), "medieval16", ypos);

				break;

			case 9:
				tvgfx_rect (0, 0, XSize*8-1, YSize-1, TVGFX_WHITE, TVGFX_TRANS);
				TVGFX_SETFONT (waukeganeb26);
				tvgfx_puts(140, 3,"FONTS 3");

				ypos = 30;
				ypos = fontdemo (FAR(t16), "t16", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(tangle16), "tangle16", ypos);
				ypos += 5;
				ypos = fontdemo (FAR(sun22), "sun22", ypos);

				break;

			default:
				// test chart

				for (int y = 0; y<226; y+=25) {
					for (int x = 0; x<416; x+=32) {
						tvgfx_rect (x, y, 32, 25, TVGFX_WHITE, TVGFX_TRANS);
					}
				}

				tvgfx_rect (0, 0, XSize*8-1, YSize-1, TVGFX_WHITE, TVGFX_TRANS);
				tvgfx_line (0, 0, XSize*8-1, YSize-1, TVGFX_WHITE);
				tvgfx_line (0, YSize-1, XSize*8-1, 0, TVGFX_WHITE);

				//tvgfx_linex (20, 20, 40, 200, TVGFX_WHITE, 5);
				tvgfx_rect (128, 25, 160, 25, TVGFX_WHITE, TVGFX_WHITE);

				//tvgfx_rect (5, 90, YSize, 49, TVGFX_BLACK, TVGFX_BLACK);
				//tvgfx_rect (300, 0, 83, YSize, TVGFX_BLACK, TVGFX_BLACK);

				tvgfx_circle (XSize*4, YSize/2, YSize/2-5, TVGFX_WHITE, TVGFX_TRANS); 

				// hor stripe pattern
				for (int i = 128; i<288; i += 2) {
					tvgfx_line (i, 75, i, 150, TVGFX_WHITE);
				}
				// ver stripe pattern
				for (int i = 0; i<YSize; i += 2) {
					tvgfx_line (288, i, 352, i, TVGFX_WHITE);
				}
				//moiree
				for (int i = 128; i<288; i+=3) {
					tvgfx_line (i, 150, XSize*4, YSize-1, TVGFX_WHITE);
				}

				TVGFX_SETFONT (terminus16);
				tvgfx_puts(XSize * 4 - 75, YSize / 2 - 8,"TEXT OUTPUT TEST");

				demo = -1;
		}
		demo++;

		_delay_ms(500);
	}

#else
	// single test image, then wait forever

	tvgfx_clr();

#if 0
	tvgfx_rect (  0, 0, 102, YSize, TVGFX_WHITE, TVGFX_WHITE);
	tvgfx_rect (204, 0, 102, YSize, TVGFX_WHITE, TVGFX_WHITE);

	//TVGFX_SETFONT (terminus16);
	//tvgfx_puts(20, 100,"TEXT OUTPUT TEST");
#else

	TVGFX_SETFONT (waukeganeb34);
	tvgfx_puts(150, 100, "C   O   M");
	TVGFX_SETFONT (waukeganeb26);
	tvgfx_puts (50, 50, "HTT 6V-12");

#endif
	while (1);

#endif

}
