/*
 Copyright (c) 2012, 2013 G. Bartsch

 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
 files (the "Software"), to deal in the Software without
 restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following
 conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.

 Credits go to:

 arduino-tvout Copyright (c) 2010 Myles Metzer 
 http://code.google.com/p/arduino-tvout/source/browse/TVout.cpp

 AVR Videogenerator, 40x25 Zeichen, nur 60% CPU Auslastung ! by Benedikt
 http://www.mikrocontroller.net/topic/53141#413260

 Simple VGA/Video Adapter von Ibragimov Maxim Rafikovich
 http://www.serasidis.gr/circuits/AVR_VGA/avr_vga.htm

 The Beauty of Bresenham's Algorithm
 A simple implementation to plot lines, circles, ellipses and Bézier curves.
 http://free.pages.at/easyfilter/bresenham.html
*/

#ifdef AVR
#include <avr/pgmspace.h>
#endif

#include <inttypes.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "video.h"
#include "draw2d.h"
//#include "dbg.h"

#define PI 3.14159

// size of segments for arcs/beziers
#define SEG_SIZE 5

void bzero(void *s, size_t n);

void tvgfx_clr(void) {
	bzero ((void *) image_mem, XSize * YSize);
}


/*
 * Inline version of tvgfx_setpixel that does not perform a bounds check
 */
static void inline sp(int16_t x, int16_t y, char c) {

	int16_t adr = (x/8) + (y*XSize);

	switch (c) {
	case TVGFX_WHITE:
		image_mem[adr] |= 0x80 >> (x&7);
		break;
	case TVGFX_BLACK:
		image_mem[adr] &= ~0x80 >> (x&7);
		break;
	case TVGFX_INVERT:
		image_mem[adr] ^= 0x80 >> (x&7);
		break;
	}

} 

void tvgfx_set_pixel(int16_t x, int16_t y, char c) {
	if (x >= XSize*8 || y >= YSize)
		return;
	sp(x,y,c);
} 

/*
 * Inline version of tvgfx_get_pixel that does not perform a bounds check
 */
static char inline gp(int16_t x, int16_t y) {
	if (image_mem[x/8+y*XSize] & (0x80 >>(x&7)))
		return 1;
	return 0;
} 

char tvgfx_get_pixel(int16_t x, int16_t y) {
	if (x >= XSize*8 || y >= YSize)
		return 0;
	if (image_mem[x/8+y*XSize] & (0x80 >>(x&7)))
		return 1;
	return 0;
} 

void tvgfx_setpixel(int16_t x, int16_t y, char c) {
	if (x >= XSize*8 || y >= YSize)
		return;
	sp(x,y,c);
} 

void tvgfx_line(int16_t x0, int16_t y0, int16_t x1, int16_t y1, char c) {

	if (x0 > XSize*8 || y0 > YSize || x1 > XSize*8 || y1 > YSize)
		return;
	if (x0 == x1)
		tvgfx_column(x0,y0,y1,c);
	else if (y0 == y1)
		tvgfx_row(y0,x0,x1,c);
	else {
		int e;
		signed int dx,dy,j, temp;
		signed char s1,s2, xchange;
		signed int x,y;

		x = x0;
		y = y0;
	
		//take absolute value
		if (x1 < x0) {
			dx = x0 - x1;
			s1 = -1;
		}
		else if (x1 == x0) {
			dx = 0;
			s1 = 0;
		}
		else {
			dx = x1 - x0;
			s1 = 1;
		}

		if (y1 < y0) {
			dy = y0 - y1;
			s2 = -1;
		}
		else if (y1 == y0) {
			dy = 0;
			s2 = 0;
		}
		else {
			dy = y1 - y0;
			s2 = 1;
		}

		xchange = 0;   

		if (dy>dx) {
			temp = dx;
			dx = dy;
			dy = temp;
			xchange = 1;
		} 

		e = ((int)dy<<1) - dx;  
	 
		for (j=0; j<=dx; j++) {
			sp(x,y,c);
		 
			if (e>=0) {
				if (xchange==1) x = x + s1;
				else y = y + s2;
				e = e - ((int)dx<<1);
			}
			if (xchange==1)
				y = y + s2;
			else
				x = x + s1;
			e = e + ((int)dy<<1);
		}
	}
} 


void tvgfx_row(int16_t y0, int16_t x0, int16_t x1, char c) {
	uint8_t lbit, rbit;

	//LDEBUG ("tvgfx_row: y0=%d, x0=%d, x1=%d\r\n", y0, x0, x1);
	
	if (x0 == x1)
		tvgfx_set_pixel(x0,y0,c);
	else {
		if (x0 > x1) {
			int16_t temp = x0;
			x0 = x1;
			x1 = temp;
		}
		lbit = 0xff >> (x0 & 7);
		x0 = x0/8 + XSize*y0;
		rbit = ~(0xff >> (x1 & 7));
		x1 = x1/8 + XSize*y0;
		if (x0 == x1) {
			lbit = lbit & rbit;
			rbit = 0;
		}
		if (c == TVGFX_WHITE) {
			image_mem[x0++] |= lbit;
			while (x0 < x1)
				image_mem[x0++] = 0xff;
			image_mem[x0] |= rbit;
		}
		else if (c == TVGFX_BLACK) {
			image_mem[x0++] &= ~lbit;
			while (x0 < x1)
				image_mem[x0++] = 0;
			image_mem[x0] &= ~rbit;
		}
		else if (c == TVGFX_INVERT) {
			image_mem[x0++] ^= lbit;
			while (x0 < x1)
				image_mem[x0++] ^= 0xff;
			image_mem[x0] ^= rbit;
		}
	}
} 


void tvgfx_column(int16_t x0, int16_t y0, int16_t y1, char c) {

	unsigned char bit;
	int byte;
	
	if (y0 == y1)
		tvgfx_set_pixel(x0,y0,c);
	else {
		if (y1 < y0) {
			bit = y0;
			y0 = y1;
			y1 = bit;
		}
		bit = 0x80 >> (x0&7);
		byte = x0/8 + y0*XSize;
		if (c == TVGFX_WHITE) {
			while ( y0 <= y1) {
				image_mem[byte] |= bit;
				byte += XSize;
				y0++;
			}
		}
		else if (c == TVGFX_BLACK) {
			while ( y0 <= y1) {
				image_mem[byte] &= ~bit;
				byte += XSize;
				y0++;
			}
		}
		else if (c == TVGFX_INVERT) {
			while ( y0 <= y1) {
				image_mem[byte] ^= bit;
				byte += XSize;
				y0++;
			}
		}
	}
}

void tvgfx_rect(int16_t x0, int16_t y0, int16_t w, int16_t h, char c, char fc) {
	
	if (fc != TVGFX_TRANS) {
		for (int16_t i = y0; i < y0+h; i++) {
			tvgfx_row(i,x0,x0+w,fc);
		}
	}
	tvgfx_line(x0,y0,x0+w,y0,c);
	tvgfx_line(x0,y0,x0,y0+h,c);
	tvgfx_line(x0+w,y0,x0+w,y0+h,c);
	tvgfx_line(x0,y0+h,x0+w,y0+h,c);
} 

void tvgfx_circle(int16_t x0, int16_t y0, int16_t radius, char c, char fc) {

	int16_t f = 1 - radius;
	int16_t ddF_x = 1;
	int16_t ddF_y = -2 * radius;
	int16_t x = 0;
	int16_t y = radius;
	int16_t pyy = y,pyx = x;


	//there is a fill color
	if (fc != TVGFX_TRANS)
		tvgfx_row(y0,x0-radius,x0+radius,fc);

	sp(x0, y0 + radius,c);
	sp(x0, y0 - radius,c);
	sp(x0 + radius, y0,c);
	sp(x0 - radius, y0,c);

	while(x < y) {
		if(f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;

		//there is a fill color
		if (fc != TVGFX_TRANS) {
			//prevent double draws on the same rows
			if (pyy != y) {
				tvgfx_row(y0+y,x0-x,x0+x,fc);
				tvgfx_row(y0-y,x0-x,x0+x,fc);
			}
			if (pyx != x && x != y) {
				tvgfx_row(y0+x,x0-y,x0+y,fc);
				tvgfx_row(y0-x,x0-y,x0+y,fc);
			}
			pyy = y;
			pyx = x;
		}
		sp(x0 + x, y0 + y,c);
		sp(x0 - x, y0 + y,c);
		sp(x0 + x, y0 - y,c);
		sp(x0 - x, y0 - y,c);
		sp(x0 + y, y0 + x,c);
		sp(x0 - y, y0 + x,c);
		sp(x0 + y, y0 - x,c);
		sp(x0 - y, y0 - x,c);
	}
} 

void tvgfx_ellipse (int16_t x0_, int16_t y0_, int16_t xr_, int16_t yr_, char c, char color_bg) {

	int32_t x0=x0_;
	int32_t y0=y0_;
	int32_t xr=xr_;
	int32_t yr=yr_;

	int32_t x = -xr, y = 0; /* II. quadrant from bottom left to top right */
	int32_t e2 = yr, dx = (1+2*x)*e2*e2;              /* error increment  */
	int32_t dy = x*x, err = dx+dy;                   /* error of 1.step */

	//LDEBUG ("ellipse: x=%ld, e2=%ld, dx=%ld, dy=%ld\r\n", x, e2, dx, dy);
	do {
		if (color_bg != TVGFX_TRANS) {
			tvgfx_row (y0+y, x0+x, x0-x, color_bg);
			//LDEBUG ("ellipse: row y=%ld, x0=%ld, x1=%ld\r\n", y0+y, x0+x, x0-x);
			tvgfx_row (y0-y, x0+x, x0-x, color_bg);
		}

		sp(x0-x, y0+y, c);                        /*   I. Quadrant */
		sp(x0+x, y0+y, c);                        /*  II. Quadrant */
		sp(x0+x, y0-y, c);                        /* III. Quadrant */
		sp(x0-x, y0-y, c);                        /*  IV. Quadrant */

		e2 = 2*err;
		if (e2 >= dx) { x++; err += dx += 2*yr*yr; }    /* x step */
		if (e2 <= dy) { y++; err += dy += 2*xr*xr; }    /* y step */

		//LDEBUG ("ellipse: x=%ld, e2=%ld, dx=%ld, dy=%ld\r\n", x, e2, dx, dy);

	} while (x <= 0);

	while (y++ < yr) {   /* too early stop for flat ellipses with xr=1, */
		sp(x0, y0+y, c);               /* -> finish tip of ellipse */
		sp(x0, y0-y, c); 
	}
}

/*
 * slower but advanced, non-bresenham code starts here
 */

void tvgfx_linex(int16_t x0, int16_t y0, int16_t x1, int16_t y1, char c, uint8_t thickness) {

	//LDEBUG ("linex: %d/%d -> %d/%d\r\n", x0, y0, x1, y1);

	tvgfx_line (x0, y0, x1, y1, c);

	if (thickness > 1) {
		int16_t dx = abs(x1 - x0);
		int16_t dy = abs(y1 - y0);

		if ( dx > dy ) {

			//float wy = (thickness-1)  *  sqrt( dx*dx + dy*dy + 0.01 )  /  (2 * dx) ;
			////LDEBUG ("linex: wy=%d, wx=%d, dx=%d, dy=%d\r\n", (int32_t) (wy * 1000), (int32_t) (wx * 1000), dx, dy);
			//LDEBUG ("linex: wy=%f, dx=%d, dy=%d\r\n", wy, dx, dy);

			int16_t wy = thickness  *  ( dx*dx + dy*dy )  /  (2 * dx * dx + 1) ;
			//LDEBUG ("linex: wy=%d, dx=%d, dy=%d\r\n", wy, dx, dy);

			for (int16_t i = 1; i < wy; i++) {
				//LDEBUG ("linex: wy i=%d, line %d/%d -> %d/%d\r\n", i, x0, y0 - i, x1, y1 - i);
				//LDEBUG ("                line %d/%d -> %d/%d\r\n", x0, y0 + i, x1, y1 + i);
				tvgfx_line(x0, y0 - i, x1, y1 - i, c);
				tvgfx_line(x0, y0 + i, x1, y1 + i, c);
			}

		} else {

			//float wx = (thickness-1)  *  sqrt( dx*dx + dy*dy + 0.01 )  /  (2 * dy) ;
			int16_t wx = thickness  *  ( dx*dx + dy*dy )  /  (2 * dy * dy + 1) ;
			//LDEBUG ("linex: wx=%d, dx=%d, dy=%d\r\n", wx, dx, dy);

			for (int16_t i = 1; i < wx; i++) {
				//LDEBUG ("linex: wx i=%d\r\n", i);
				tvgfx_line(x0 - i, y0, x1 - i, y1, c);
				tvgfx_line(x0 + i, y0, x1 + i, y1, c);
			}
		}
	}
}

void tvgfx_rectx(int16_t x0, int16_t y0, int16_t w, int16_t h, char c, char color_bg, uint8_t thickness) {
	
	if (color_bg != TVGFX_TRANS) {
		for (int16_t i = y0; i < y0+h; i++) {
			tvgfx_row (i, x0, x0+w, color_bg);
		}
	}

	tvgfx_linex (x0, y0, x0+w, y0, c, thickness);
	tvgfx_linex (x0, y0, x0, y0+h, c, thickness);
	tvgfx_linex (x0+w, y0, x0+w, y0+h, c, thickness);
	tvgfx_linex (x0, y0+h, x0+w, y0+h, c, thickness);
} 

/**************************************************************
 *
 *
 * 2D Engine
 *
 *
 **************************************************************/

static uint8_t engine_fg = TVGFX_WHITE;
static uint8_t engine_bg = TVGFX_WHITE;
static uint8_t engine_thickness = 3;

#define MAX_POLY_NODES 255
static int16_t nodes[MAX_POLY_NODES][2];
static uint8_t nodes_n = 0;

void tvgfx_set_thickness (uint8_t thickness) {
	engine_thickness = thickness;
}
uint8_t tvgfx_get_thickness (void) {
	return engine_thickness;
}
void tvgfx_set_foreground (uint8_t color) {
	engine_fg = color;
}
uint8_t tvgfx_get_foreground (void) {
	return engine_fg;
}
void tvgfx_set_background (uint8_t color) {
	engine_bg = color;
}
uint8_t tvgfx_get_background (void) {
	return engine_bg;
}

void tvgfx_move_to (int16_t x, int16_t y) {
	nodes_n = 0;
	tvgfx_line_to (x, y);
}
void tvgfx_line_to (int16_t x, int16_t y) {
	if (nodes_n >= MAX_POLY_NODES-1)
		return;
	nodes[nodes_n][0] = x;
	nodes[nodes_n][1] = y;
	nodes_n++;
}

static float vec_angle (float ux, float uy, float vx, float vy) {


	float a = acosf ( ( ux*vx + uy*vy) / (sqrt (ux*ux+uy*uy) * sqrt(vx*vx+vy*vy)) );

	if ((ux*vy - uy*vx) < 0)
		a = -a;

	//LDEBUG ("vec_angle: u=%f/%f, v=%f/%f => %f\r\n", ux,uy, vx, vy, a * 180.0 / PI);

	return a;
}

/*
 * Conversion from endpoint to center parameterization
 *
 * http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
 */
void tvgfx_arc_to (int16_t x2_, int16_t y2_, char fa, char fs, int16_t rx_, int16_t ry_, int16_t phi_) {

	if (!nodes_n) {
		//LDEBUG ("arc_to: ERR, no path\r\n");
		return;
	}

	float x1 = nodes[nodes_n-1][0];
	float y1 = nodes[nodes_n-1][1];
	float x2 = x2_;
	float y2 = y2_;
	float rx = rx_;
	float ry = ry_;
	float phi = PI * ((float) phi_) / 180.0;

	//LDEBUG ("arc_to: x1=%f, y1=%f, x2=%f, y2=%f, phi=%f\r\n", x1, y1, x2, y2, phi);

	// F.6.6 Correction of out-of-range radii
	
	// Step 1: Ensure radii are non-zero

	if (!rx_ || !ry_) {
		// straight line
		tvgfx_line_to (x2_, y2_);
	}

	// Step 2: Ensure radii are positive

	if (rx<0) rx = -rx;
	if (ry<0) ry = -ry;

	// F.6.5 Conversion from endpoint to center parameterization
	
	// Step 1: Compute (x1′, y1′)
	
	float x1s =  cos(phi) * ((x1-x2)/2) + sin(phi) * ((y1-y2)/2);
	float y1s = -sin(phi) * ((x1-x2)/2) + cos(phi) * ((y1-y2)/2);

	//LDEBUG ("arc_to: x1s=%f, y1s=%f\r\n", x1s, y1s);

	// F.6.6 Step 3: Ensure radii are large enough

	float l = (x1s*x1s) / (rx*rx) + (y1s*y1s) / (ry*ry);
	if (l>1) {
		rx = sqrt(l) * rx ;
		ry = sqrt(l) * ry ;

		//LDEBUG ("arc_to: radii too small, corrected to: %f/%f\r\n", rx, ry);
	}

	// Step 2: Compute (cx′, cy′)
	
	float oben =  rx*rx*ry*ry - rx*rx*y1s*y1s - ry*ry*x1s*x1s;
	if (oben <0) oben = 0; // rounding err

	float wurzel = sqrt ( oben / (rx*rx*y1s*y1s + ry*ry*x1s*x1s) );

	if (fa == fs) {
		wurzel = - wurzel; 
	}

	float cxs = wurzel * (rx*y1s/ry);
	float cys = wurzel * (-ry*x1s/rx);

	//LDEBUG ("arc_to: cxs=%f, cys=%f, wurzel=%f, oben=%f\r\n", cxs, cys, wurzel, oben);

	// Step 3: Compute (cx, cy) from (cx′, cy′)
	
	float cx = cos(phi)*cxs - sin(phi)*cys + (x1+x2)/2;
	float cy = sin(phi)*cxs + cos(phi)*cys + (y1+y2)/2;

	//LDEBUG ("arc_to: cx=%f, cy=%f\r\n", cx, cy);

	// Step 4: Compute θ1 and Δθ
	
	float t1 = vec_angle ( 1, 0, (x1s-cxs)/rx, (y1s-cys)/ry );
	float dt = vec_angle ( (x1s-cxs)/rx, (y1s-cys)/ry, (-x1s-cxs)/rx, (-y1s-cys)/ry);

	if (fs) {
		if (dt < 0) dt += 2*PI;
	} else {
		if (dt > 0) dt -= 2*PI;
	}

	float t2 = t1 + dt;

	if (t2<t1) {
		float temp = t1;
		t1 = t2;
		t2 = temp;
	}

	//LDEBUG ("arc_to: t1=%f, dt=%f, t2=%f\r\n", t1 * 180.0 / PI, dt * 180.0 / PI, t2 * 180.0 / PI);

	/*
	 * compute arc nodes
	 */

	float t = t1;
	float step = 2 * PI / SEG_SIZE;

	while (1) {

		int16_t x= cx + rx * cos(t) * cos (phi) - ry * sin(t) * sin(phi);
		int16_t y= cy + rx * cos(t) * sin (phi) + ry * sin(t) * cos(phi);

		if (t > t1) {

			//LDEBUG ("arc: line_to %d/%d\r\n", x, y);

			tvgfx_line_to (x, y);
		}

		if (t == t2) 
			break;

		t += step;
		if (t > t2)
			t = t2;
	}
}

void tvgfx_qbezier_to (int16_t x1, int16_t y1, int16_t x2, int16_t y2) {

	if (!nodes_n) {
		//LDEBUG ("qbezier_to: ERR, no path\r\n");
		return;
	}

	int16_t x0 = nodes[nodes_n-1][0];
	int16_t y0 = nodes[nodes_n-1][1];

	// compute number of segments (depends on length of curve)

	float dx = x2 - x0;
	float dy = y2 - y0;

	int16_t n_seg =  sqrt (dx*dx + dy*dy) / SEG_SIZE + 1;

	//LDEBUG ("qbezier: nseg=%d\r\n", n_seg);

	int16_t x, y;
	for (uint8_t i=1; i <= n_seg; ++i) {
		float t = (float)i / (float)n_seg;
		float a = pow((1.0 - t), 2.0);
		float b = 2.0 * t * (1.0 - t);
		float c = pow(t, 2.0);
		x = (int16_t) (a * x0 + b * x1 + c * x2);
		y = (int16_t) (a * y0 + b * y1 + c * y2);

		tvgfx_line_to (x, y);
	}
}

void tvgfx_cbezier_to (int16_t x1, int16_t y1, int16_t x2, int16_t y2, int16_t x3, int16_t y3) {
	if (!nodes_n) {
		//LDEBUG ("cbezier_to: ERR, no path\r\n");
		return;
	}

	int16_t x0 = nodes[nodes_n-1][0];
	int16_t y0 = nodes[nodes_n-1][1];

	// compute number of segments (depends on length of curve)

	float dx = x3 - x0;
	float dy = y3 - y0;

	int16_t n_seg =  sqrt (dx*dx + dy*dy) / SEG_SIZE + 1;

	//LDEBUG ("cbezier: nseg=%d\r\n", n_seg);

	int16_t x, y;

	for (uint8_t i=1; i <= n_seg; ++i) {
		float t = (float)i / (float) n_seg;

		float a = pow((1.0 - t), 3.0);
		float b = 3.0 * t * pow((1.0 - t), 2.0);
		float c = 3.0 * pow(t, 2.0) * (1.0 - t);
		float d = pow(t, 3.0);

		x = (int16_t) (a * x0 + b * x1 + c * x2 + d * x3);
		y = (int16_t) (a * y0 + b * y1 + c * y2 + d * y3);

		tvgfx_line_to (x, y);
	}
}

void tvgfx_fill (void) {

	int16_t xi[MAX_POLY_NODES];
	float slope[MAX_POLY_NODES];

	nodes[nodes_n][0]=nodes[0][0];
	nodes[nodes_n][1]=nodes[0][1];

	if (engine_bg != TVGFX_TRANS) {
		for (uint8_t i=0; i<nodes_n; i++)
		{
			int16_t dy = nodes[i+1][1] - nodes[i][1];
			int16_t dx = nodes[i+1][0] - nodes[i][0];

			if (dy==0) slope[i] = 1.0;
			if (dx==0) slope[i] = 0.0;

			if ((dy!=0) && (dx!=0)) { // calculate inverse slope 
				slope[i] = (float) dx/dy;
			}
		}

		for (int16_t y=0; y<YSize; y++) {
			uint8_t k = 0;
			for (uint8_t i=0; i<nodes_n; i++) {

				if( ((nodes[i][1]<=y)&&(nodes[i+1][1]>y))||
						((nodes[i][1]>y)&&(nodes[i+1][1]<=y))) {
					xi[k] = (int16_t) (nodes[i][0] + slope[i]*(y-nodes[i][1]));
					k++;
				}
			}

			for (uint8_t j=0; j<k-1; j++) { // Arrange x-intersections in order 
				for (uint8_t i=0;i<k-1;i++) {
					if (xi[i] > xi[i+1]) {
						int16_t temp = xi[i];
						xi[i] = xi[i+1];
						xi[i+1] = temp;
					}
				}
			}

			for (uint8_t i=0; i<k; i+=2) {
				tvgfx_line (xi[i], y, xi[i+1]+1, y, engine_bg);
			}

		}
	}

	/* draw polygon outline */

	for (uint8_t i=0; i<nodes_n-1; i++) {
		//LDEBUG ("draw2d: path_close: poly node #%3d: %d/%d\r\n", i, nodes[i][0], nodes[i][1]);
		tvgfx_linex (nodes[i][0], nodes[i][1], nodes[i+1][0], nodes[i+1][1], engine_fg, engine_thickness);
	}
	nodes_n = 0;
}

void tvgfx_stroke (void) {

	/* draw polygon outline */

	for (uint8_t i=0; i<nodes_n-1; i++) {
		//LDEBUG ("draw2d: path_close: poly node #%3d: %d/%d\r\n", i, nodes[i][0], nodes[i][1]);
		tvgfx_linex (nodes[i][0], nodes[i][1], nodes[i+1][0], nodes[i+1][1], engine_fg, engine_thickness);
	}

	nodes_n = 0;
}


/*
 * flood fill
 */

#define FILL_STACK_SIZE 32
static int16_t stack[FILL_STACK_SIZE];

static uint8_t stack_pointer = 0;

static char pop(int16_t *x, int16_t *y) { 

	if (stack_pointer > 1) { 

		*x = stack[--stack_pointer];
		*y = stack[--stack_pointer];

		return 1; 

	} else { 

		return 0; 
	}    
}    

static char push(int16_t x, int16_t y) { 

	if (stack_pointer < FILL_STACK_SIZE - 2) { 

		stack[stack_pointer++] = y; 
		stack[stack_pointer++] = x; 

		return 1; 

	} else { 

		return 0; 

	}    
}     

static void emptyStack(void) { 
	int16_t x, y; 
	while (pop(&x, &y)); 
}


void tvgfx_flood (int16_t x, int16_t y) {

	//LDEBUG ("draw2d: fill at %d/%d\r\n", x, y);

	emptyStack();

	if (!push(x, y)) 
		return; 
	int16_t y1; 
	char spanLeft, spanRight;
	while (pop(&x, &y)) {
		y1 = y;
		while(y1 >= 0 && !gp(x, y1) ) y1--;
		y1++;
		spanLeft = spanRight = 0;
		while(y1 < YSize && !gp(x, y1) )
		{
			sp (x, y1, TVGFX_WHITE);
			if(!spanLeft && x > 0 && !gp(x - 1, y1)) 
			{
				if(!push(x - 1, y1)) return;
				spanLeft = 1;
			}
			else if(spanLeft && x > 0 && gp(x - 1, y1))
			{
				spanLeft = 0;
			}
			if(!spanRight && x < XSize*8 - 1 && !gp(x + 1, y1)) 
			{
				if(!push(x + 1, y1)) return;
				spanRight = 1;
			}
			else if(spanRight && x < XSize*8 - 1 && gp(x + 1, y1))
			{
				spanRight = 0;
			} 
			y1++;
		}
	}     
}

/*
 *
 * bitmaps, fonts
 *
 */

#ifdef TVGFX_FASTFONTS
void tvgfx_bitmap(int16_t x, int16_t y, const unsigned char *bmp, int16_t offset, uint8_t width, uint8_t lines) {
#else
void tvgfx_bitmap(int16_t x, int16_t y, uint32_t bmp, int16_t offset, uint8_t width, uint8_t lines) {
#endif

#ifdef AVR

	// crude clipping
	if (x + width >= XSize*8)
		return;

        uint8_t temp, lshift, rshift, save, xtra;
        int16_t si = 0;
        
        rshift = x&7;
        lshift = 8-rshift;
                
        if (width&7) {
                xtra = width&7;
                width = width/8;
                width++;
        } else {
                xtra = 8;
                width = width/8;
        }
        
        for (uint8_t l = 0; l < lines; l++) {
                si = (y + l)*XSize + x/8;
                if (width == 1)
                        temp = 0xff >> (rshift + xtra);
                else
                        temp = 0;
                save = image_mem[si];
                image_mem[si] &= ((0xff << lshift) | temp);
#ifdef TVGFX_FASTFONTS
                temp = pgm_read_byte ((uint16_t)bmp + (offset++));
#else
                temp = pgm_read_byte_far (bmp + offset++);
#endif
                image_mem[si++] |= temp >> rshift;
                for ( int16_t b = offset + width-1; offset < b; offset++) {
                        save = image_mem[si];
                        image_mem[si] = temp << lshift;
#ifdef TVGFX_FASTFONTS
                        temp = pgm_read_byte ((uint16_t)bmp + offset);
#else
                        temp = pgm_read_byte_far (bmp + offset);
#endif

                        image_mem[si++] |= temp >> rshift;
                }
                if (rshift + xtra < 8)
                        image_mem[si-1] |= (save & (0xff >> (rshift + xtra)));
                if (rshift + xtra - 8 > 0)
                        image_mem[si] &= (0xff >> (rshift + xtra - 8));
                image_mem[si] |= temp << lshift;
        }
#endif
} 


#ifdef AVR
// current font + cached information about it:

static int16_t font_chpx = 0;  // char height in pixels
static int16_t font_cwpx;      // char width (0 for proportional fonts)
static int16_t font_offs[128]; // offset of each character
#endif

#ifdef TVGFX_FASTFONTS
static const unsigned char *font;
void tvgfx_setfont (const unsigned char *f) {
#else
static uint32_t font;
void tvgfx_setfont (uint32_t f) {
#endif
#ifdef AVR
	int16_t eoff;  // encoding offset

	//LDEBUG ("tvgfx_setfont: font=%8lx\r\n", f);

	font = f;

#ifdef TVGFX_FASTFONTS
	font_chpx = pgm_read_byte((uint16_t)font+1);
	font_cwpx = pgm_read_byte((uint16_t)font);
	eoff = pgm_read_byte((uint16_t)font+2);
#else
	font_chpx = pgm_read_byte_far(font+1);
	font_cwpx = pgm_read_byte_far(font);
	eoff = pgm_read_byte_far(font+2);
#endif
	//LDEBUG ("tvgfx_setfont: chpx=%d\r\n", font_chpx);
	//LDEBUG ("tvgfx_setfont: cwpx=%d\r\n", font_cwpx);
	//LDEBUG ("tvgfx_setfont: eoff=%d\r\n", eoff);

 	if (font_cwpx) {
		int16_t cw;    // width in bytes
		cw = (font_cwpx + 7) / 8; 
		for (int16_t i = 0; i<127; i++) {
			if (i<eoff) {
				font_offs[i] = 3;
			} else {
				font_offs[i] = (i-eoff) * cw * font_chpx + 3;
			}
		}
	} else {
		int16_t off=3; // offset into font

		for (int16_t i = 0; i<127; i++) {
			if (i<eoff) {
				font_offs[i] = 3;
			} else {

#ifdef TVGFX_FASTFONTS
				uint8_t cwpx = pgm_read_byte((uint16_t)font+off);
#else
				uint8_t cwpx = pgm_read_byte_far(font+off);
#endif
				font_offs[i] = off;
				off += ((cwpx + 7) / 8) * font_chpx + 1; 
			}
		}
	}
#endif
}

uint8_t tvgfx_putc(int16_t x, int16_t y, unsigned char c) {
#ifdef AVR
	int16_t off = font_offs[c];
	uint8_t w = font_cwpx;

	if (!w) {
#ifdef TVGFX_FASTFONTS
		w = pgm_read_byte((uint16_t)font+off);
#else
		w = pgm_read_byte_far (font+off);
#endif
		off++;
	}

	tvgfx_bitmap (x, y, font, off, w, font_chpx);

	return w;
#else
	return 0;
#endif
}

int16_t tvgfx_puts (int16_t x, int16_t y, char *str) {

	int16_t width = 0;

	while (*str) {
		uint8_t w = tvgfx_putc (x, y, *str++);
		x += w;
		width += w;
	}

	return width;
}

uint8_t tvgfx_getfontheight(void) {
#ifdef AVR
	return font_chpx;
#else 
	return 0;
#endif
}

static uint8_t tvgfx_charwidth(unsigned char c) {

#ifdef AVR
	int16_t off = font_offs[c];
	uint8_t w = font_cwpx;

	if (!w) {
#ifdef TVGFX_FASTFONTS
		w = pgm_read_byte((uint16_t)font+off);
#else
		w = pgm_read_byte_far (font+off);
#endif
	}

	return w;
#else
	return 0;
#endif
}

int16_t tvgfx_textwidth(char *str) {
	int16_t w = 0;
	while (*str) {
		w += tvgfx_charwidth (*str++);
	}
	return w;
}

void tvgfx_printf(int16_t x, int16_t y, const char *fmt, ...) {

        va_list ap;
        char buf[128];

        va_start(ap, fmt);
        vsnprintf (buf, 128, fmt, ap);
        va_end(ap);

        tvgfx_puts(x, y, buf);

}
