/*
 Copyright (c) 2012, 2013 G. Bartsch

 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
 files (the "Software"), to deal in the Software without
 restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following
 conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.

 Credits go to:

 arduino-tvout Copyright (c) 2010 Myles Metzer 
 http://code.google.com/p/arduino-tvout/source/browse/TVout.cpp

 AVR Videogenerator, 40x25 Zeichen, nur 60% CPU Auslastung ! by Benedikt
 http://www.mikrocontroller.net/topic/53141#413260

 Simple VGA/Video Adapter von Ibragimov Maxim Rafikovich
 http://www.serasidis.gr/circuits/AVR_VGA/avr_vga.htm
*/

#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>

#include "video.h"

// internal line counter
static uint16_t m_Y_Cnt=0;

#define FPSize      (VLines-YSize)-VSync-BPSize    // front porch in lines
#define FPStart     YSize                          // m_Y_Cnt when front porch starts
#define VSStart     FPStart + FPSize               // m_Y_Cnt when Vsync starts
#define BPStart     VSStart + VSync                // m_Y_cnt when Backporch starts 

volatile char image_mem[XSize * YSize];

static void (*hbi_hook)(void) = NULL;
static void (*vbi_hook)(void) = NULL;

void bzero(void *s, size_t n);

void tvgfx_init(void) {

	// UART1 in SPI mode

	UBRR1 = 0;
	UCSR1C = (1<<UMSEL11)|(1<<UMSEL10);
	UCSR1B = (0<<RXEN1)|(1<<TXEN1);
	UBRR1 = 0;

        // Timer 1 (tv output)

        ICR1 = 1024;

        OCR1A = HSync;
        OCR1B = XLineTimer;      // triggers line content output
        TCCR1A = (1<<COM1A1)|(1<<COM1A0)|(1<<WGM11)|(1<<WGM10);
        TCCR1B =  (1<<WGM12)|1 ; // 10bit PWM at 16MHz: 15,625kHz
        TIMSK1 = 1<<OCIE1B;

        bzero ((void *) image_mem, XSize * YSize);
}

// very useful assembly code to wait cycle-exact for timer1 to reach a certain value
// credits go to arduino-tvout for this!
static void inline wait_until(uint8_t time) {
        __asm__ __volatile__ (
                        "subi   %[time], 10\n"
                        "sub    %[time], %[tcnt1l]\n\t"
                "100:\n\t"
                        "subi   %[time], 3\n\t"
                        "brcc   100b\n\t"
                        "subi   %[time], 0-3\n\t"
                        "breq   101f\n\t"
                        "dec    %[time]\n\t"
                        "breq   102f\n\t"
                        "rjmp   102f\n"
                "101:\n\t"
                        "nop\n" 
                "102:\n"
                :
                : [time] "a" (time),
                [tcnt1l] "a" (TCNT1L)
        );
}

ISR(TIMER1_COMPB_vect)
{
	if (m_Y_Cnt < YSize) {

		// compensate for interrupt latency

		UDR1 = 0;
		wait_until (XLineTimer + XLineDelay);

		// start output of line content via SPI
	
		volatile char *imgptr = image_mem + m_Y_Cnt * XSize ;

		uint8_t px = *imgptr++;

		UCSR1B = (0<<RXEN1)|(1<<TXEN1);

		// fill transmit buffer with first data byte
		UDR1 = px;
		px = *imgptr++;

		for (uint16_t i = 1; i<XSize; i++) {

			/* Wait for empty transmit buffer */
			while ( !( UCSR1A & (1<<UDRE1)) );

			/* Put data into buffer, sends the data */
			UDR1 = px;
			px = *imgptr++;

		}

		// black border

		UDR1 = 0;
		UCSR1B = (0<<RXEN1)|(0<<TXEN1);

		asm volatile ("nop");
		asm volatile ("nop");

		m_Y_Cnt++;

	} else {

		// VSync

		m_Y_Cnt++;

		if (m_Y_Cnt == VSStart) {

			//  Sync impulse active high
			TCCR1A = (1<<COM1A1)|(0<<COM1A0)|(1<<WGM11)|(1<<WGM10);

		} else if (m_Y_Cnt == BPStart) {

			//  Sync impulse active low
			TCCR1A = (1<<COM1A1)|(1<<COM1A0)|(1<<WGM11)|(1<<WGM10);

		} else if (m_Y_Cnt >= VLines) {

			m_Y_Cnt = 0;	

			sei();
			if (vbi_hook) {
				void (*tmp)(void) = vbi_hook;
				vbi_hook = NULL;
				tmp();
			}
		}

	}

	sei();
	if (hbi_hook) {
		void (*tmp)(void) = hbi_hook;
		hbi_hook = NULL;
		tmp();
	}
}

void tvgfx_set_hbi_hook(void (*hook)(void)) {
	hbi_hook = hook;
}

void tvgfx_set_vbi_hook(void (*hook)(void)) {
	vbi_hook = hook;
}

